"""
A thin wrapper of the NetApp ONTAP API.


Generic ONTAP error codes
-------------------------

1	    An entry with the same identifiers already exists.
2	    A field has an invalid value, is missing, or an extra field was provided.
3	    The operation is not supported.
4	    An entry with the specified identifiers was not found.
6	    Permission denied.
7	    Resource limit exceeded.
8	    Resource in use.
65541	RPC timed out.
65552	Generic RPC failure.
65562	Internal RPC error
262145	Application code returned an unexpected exception.
262160	There are too many requets already being processed. Retry after a short delay.
262177	Missing value.
262179	Unexpected argument. Argument shown in error message body.
262185	Invalid value with value in the body of the error.
262186	A field is used in an invalid context with another field,
        as shown in error message body.
262188	A field was specified twice. Location of assignments shown in error message body.
262190	You must provide one or more values to apply your changes.
262196	Field cannot be set in this operation.
262197	Invalid value provided for field. Value and field shown in error message body.
262198	A request body is not allowed on GET, HEAD, and DELETE.
262199	Invalid JSON with error location provided in body of the error.
262200	Invalid JSON range, with range provided in the body of the error.
262201	Invalid JSON due to unknown formatting issue.
262202	Field is considered secret and should not be provided in the URL.
262210	Unable to retrieve all required records within the timeout.
        This "order_by" query is not supported under the current system load
        with the current number of records in the collection.
262211	POST request on a REST API does not support filtering on an attribute.
        Attributes must be in the request body.
262212	Request is missing required value.
262220	Wildcard fields=* is not allowed for CLI-based REST APIs.
262245	Invalid value with reason provided in body of the error.
262247	Invalid value for a field, with value and field in body of the error.
262248	A value is missing assignment operator.
262249	Field name is not supported for GET requests because it is not a readable attribute.
262250	Field name cannot be queried in a GET request because it is not a readable attribute.
262254	Invalid JSON input, an array was expected.
262255	An array was found in the JSON when it was not expected.
262268	The field is not supported as an order_by= field.
262277	The query_fields and query parameters may only be specified for GET requests.
262282	Property was specified twice.
262286	Mismatching braces found in the fields= query.
393271	A node is out of quorum. Body of error message identifies node.
39387137	A provided URL is invalid.
"""

from __future__ import annotations

import asyncio
import contextlib
import enum
import uuid
from collections.abc import Iterable
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    List,
    Mapping,
    NotRequired,
    Optional,
    Sequence,
    TypeAlias,
    TypedDict,
)

import aiohttp

from ..exception import ExternalError
from ..types import QuotaConfig, QuotaUsage

StorageID: TypeAlias = uuid.UUID
VolumeID: TypeAlias = uuid.UUID
QTreeID: TypeAlias = int


class JobResponseCode(enum.StrEnum):
    QUOTA_ALEADY_ENABLED = "5308507"


class AsyncJobResult(TypedDict):
    state: str
    code: str | None
    message: str | None


class SpaceInfo(TypedDict):
    available: int
    used: int
    size: int


class SVMInfo(TypedDict):
    uuid: str
    name: str


class VolumeInfo(TypedDict):
    name: str
    uuid: VolumeID
    path: Path
    space: NotRequired[SpaceInfo]
    statistics: NotRequired[dict[str, Any]]
    svm: NotRequired[SVMInfo]


class QTreeInfo(TypedDict):
    id: int
    name: str
    path: Path
    security_style: NotRequired[str]
    export_policy: NotRequired[dict[str, Any]]
    statistics: NotRequired[dict[str, Any]]


class NetAppClientError(ExternalError):
    pass


class NetAppClient:
    endpoint: str
    user: str
    password: str
    user_id: int
    group_id: int
    _session: aiohttp.ClientSession

    def __init__(
        self,
        endpoint: str,
        user: str,
        password: str,
        user_id: int,
        group_id: int,
    ) -> None:
        self.endpoint = endpoint
        self.user = user
        self.password = password
        self.user_id = user_id
        self.group_id = group_id
        _connector = aiohttp.TCPConnector(ssl=False)
        _auth = aiohttp.BasicAuth(self.user, self.password)
        self._session = aiohttp.ClientSession(
            self.endpoint,
            connector=_connector,
            auth=_auth,
            # raise_for_status=True,
        )

    async def aclose(self) -> None:
        await self._session.close()

    @contextlib.asynccontextmanager
    async def send_request(
        self,
        method: str,
        path: str,
        params: dict[str, str | int] | None = None,
        data: dict[str, Any] | None = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        async with self._session.request(method=method, url=path, params=params, json=data) as resp:
            yield resp

    async def wait_job(self, job_id: str) -> AsyncJobResult:
        while True:
            async with self.send_request("get", f"/api/cluster/jobs/{job_id}") as resp:
                data = await resp.json()
                match data["state"]:
                    case "running":
                        await asyncio.sleep(0.1)
                        continue
                    case "failure":
                        error_code = data["error"]["code"]
                        error_msg = data["error"]["message"]
                    case _:
                        error_code = None
                        error_msg = None
                return {
                    "state": data["state"],
                    "code": error_code,
                    "message": error_msg,
                }

    @staticmethod
    def check_job_result(result: AsyncJobResult, allowed_codes: Iterable[JobResponseCode]) -> None:
        _allowed_codes = {code.value for code in allowed_codes}
        if result["state"] == "failure":
            if result["code"] in _allowed_codes:
                pass
            else:
                raise NetAppClientError(f"{result['state']} [{result['code']}] {result['message']}")

    async def get_volume_metadata(self, volume_id: VolumeID) -> Mapping[str, Any]:
        raise NotImplementedError
        # qos = await self.get_qos_by_volume_id(volume_id)
        # qos_policies = await self.get_qos_policies()
        # qtree = await self.get_default_qtree(
        #     volume_id,
        #     [
        #         "security_style",
        #         "export_policy",
        #         "statistics",
        #     ],
        # )

        # # mapping certain data for better explanation
        # volume_qtree_cluster = {
        #     # ------ use volume info ------
        #     "id": data["uuid"],
        #     "local_tier": data["aggregates"][0]["name"],
        #     "create_time": data["create_time"],
        #     "snapshot_policy": data["snapshot_policy"]["name"],
        #     "snapmirroring": str(data["snapmirror"]["is_protected"]),
        #     "state": data["state"],
        #     "style": data["style"],
        #     "svm_name": data["svm"]["name"],
        #     "svm_id": data["svm"]["uuid"],
        #     # ------ use qtree info ------
        #     "name": qtree["name"],
        #     "path": qtree["path"],
        #     "security_style": qtree["security_style"],
        #     "export_policy": qtree["export_policy"]["name"],
        #     "timestamp": qtree["statistics"].get("timestamp"),  # last check time
        # }
        # # optional values to add in volume_qtree_cluster
        # if qos:
        #     volume_qtree_cluster.update({"qos": json.dumps(qos["policy"])})
        # if qos_policies:
        #     volume_qtree_cluster.update({"qos_policies": json.dumps(qos_policies)})
        # return volume_qtree_cluster

    async def list_volumes(
        self, extra_fields: Optional[Sequence[str]] = None
    ) -> Mapping[VolumeID, VolumeInfo]:
        default_extra_fields = ["nas.path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            "/api/storage/volumes",
            params={"fields": ",".join(_extra_fields)},
        ) as resp:
            data = await resp.json()
            items = {}
            for record in data["records"]:
                volume_id = VolumeID(record.pop("uuid"))
                items[volume_id] = VolumeInfo(
                    uuid=volume_id,
                    name=record.pop("name"),
                    path=Path(record.pop("nas", {}).pop("path")),
                )
                items[volume_id].update(record)
            return items

    async def get_volume_by_name(
        self,
        name: str,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> VolumeInfo:
        default_extra_fields = ["nas.path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            "/api/storage/volumes",
            params={"name": name, "fields": ",".join(_extra_fields)},
        ) as resp:
            data = await resp.json()
            record = data["records"][0]
            volume_info = VolumeInfo(
                uuid=VolumeID(record.pop("uuid")),
                name=record.pop("name"),
                path=Path(record.pop("nas", {}).pop("path")),
            )
            volume_info.update(record)
            return volume_info

    async def get_volume_by_id(
        self,
        volume_id: VolumeID,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> VolumeInfo:
        default_extra_fields = ["nas.path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            f"/api/storage/volumes/{volume_id}",
            params={"fields": ",".join(_extra_fields)},
        ) as resp:
            record = await resp.json()
            volume_info = VolumeInfo(
                uuid=VolumeID(record.pop("uuid")),
                name=record.pop("name"),
                path=Path(record.pop("nas", {}).pop("path")),
            )
            volume_info.update(record)
            return volume_info

    async def get_volume_metric_by_id(
        self,
        volume_id: VolumeID,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> dict[str, Any]:
        default_extra_fields = ["path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            f"/api/storage/volumes/{volume_id}/metrics",
            params={"fields": ",".join(_extra_fields)},
        ) as resp:
            return await resp.json()

    async def list_qtrees(
        self,
        volume_id: VolumeID,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> Sequence[QTreeInfo]:
        default_extra_fields = ["path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            f"/api/storage/qtrees/{volume_id}",
            params={"fields": ",".join(_extra_fields)},
        ) as resp:
            data = await resp.json()
            items = []
            for record in data["records"]:
                item = QTreeInfo(
                    name=record.pop("name"),
                    id=int(record.pop("id")),
                    path=Path(record.pop("path")),
                )
                item.update(record)
                items.append(item)
            return items

    async def get_default_qtree(
        self,
        volume_id: VolumeID,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> QTreeInfo:
        return await self.get_qtree_by_id(volume_id, 0, extra_fields=extra_fields)

    async def get_qtree_by_id(
        self,
        volume_id: VolumeID,
        qtree_id: QTreeID,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> QTreeInfo:
        default_extra_fields = ["path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            f"/api/storage/qtrees/{volume_id}/{qtree_id}",
            params={"fields": ",".join(_extra_fields)},
        ) as resp:
            record = await resp.json()
            qtree_info = QTreeInfo(
                name=record.pop("name"),
                id=int(record.pop("id")),
                path=Path(record.pop("path")),
            )
            qtree_info.update(record)
            return qtree_info

    async def get_qtree_by_name(
        self,
        volume_id: VolumeID,
        name: str,
        extra_fields: Optional[Sequence[str]] = None,
    ) -> QTreeInfo:
        default_extra_fields = ["path"]
        _extra_fields = (
            [*default_extra_fields, *extra_fields] if extra_fields else default_extra_fields
        )
        async with self.send_request(
            "get",
            "/api/storage/qtrees",
            params={
                "name": name,
                "volume.uuid": str(volume_id),
                "fields": ",".join(_extra_fields),
            },
        ) as resp:
            data = await resp.json()
            if data["num_records"] > 0:
                record = data["records"][0]
                qtree_info = QTreeInfo(
                    name=record.pop("name"),
                    id=int(record.pop("id")),
                    path=Path(record.pop("path")),
                )
                qtree_info.update(record)
                return qtree_info
            else:
                raise RuntimeError(f"No qtree {name} found in the volume {volume_id}")

    async def create_qtree(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
    ) -> AsyncJobResult:
        async with self.send_request(
            "post",
            "/api/storage/qtrees",
            data={
                "svm.uuid": str(svm_id),
                "volume.uuid": str(volume_id),
                "name": qtree_name,
                "user.id": self.user_id,
                "group.id": self.group_id,
            },
        ) as resp:
            data = await resp.json()
        return await self.wait_job(data["job"]["uuid"])

    async def set_quota_rule(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
        config: QuotaConfig,
    ) -> AsyncJobResult:
        async with self.send_request(
            "post",
            "/api/storage/quota/rules",
            data={
                "svm.uuid": str(svm_id),
                "volume.uuid": str(volume_id),
                "type": "tree",
                "qtree.name": qtree_name,
                "space": {
                    "hard_limit": config.limit_bytes,
                    "soft_limit": config.limit_bytes,
                },
                # 'files': {  # not supported yet from Backend.AI
                #     'hard_limit': 0,
                #     'soft_limit': 0,
                # },
            },
        ) as resp:
            data = await resp.json()
        return await self.wait_job(data["job"]["uuid"])

    async def _find_quota_rule(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
    ) -> dict[str, Any]:
        async with self.send_request(
            "get",
            "/api/storage/quota/rules",
            params={
                "svm.uuid": str(svm_id),
                "volume.uuid": str(volume_id),
                "type": "tree",
                "qtree.name": qtree_name,
                "fields": "space,files",
            },
        ) as resp:
            data = await resp.json()
            records = data["records"]
            if data["num_records"] == 0:
                raise NetAppClientError(
                    f"Quota rule not found for the volume {volume_id} and the qtree {qtree_name}"
                )
            return records[0]

    async def update_quota_rule(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
        config: QuotaConfig,
    ) -> AsyncJobResult:
        record = await self._find_quota_rule(svm_id, volume_id, qtree_name)
        async with self.send_request(
            "patch",
            f"/api/storage/quota/rules/{record['uuid']}",
            data={
                "space": {
                    "hard_limit": config.limit_bytes,
                    "soft_limit": config.limit_bytes,
                },
                # 'files': {  # not supported yet from Backend.AI
                #     'hard_limit': 0,
                #     'soft_limit': 0,
                # },
            },
        ) as resp:
            data = await resp.json()
        return await self.wait_job(data["job"]["uuid"])

    async def get_quota_rule(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
    ) -> QuotaConfig:
        record = await self._find_quota_rule(svm_id, volume_id, qtree_name)
        return QuotaConfig(
            limit_bytes=record["space"]["hard_limit"],
        )

    async def delete_quota_rule(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
    ) -> AsyncJobResult:
        record = await self._find_quota_rule(svm_id, volume_id, qtree_name)
        async with self.send_request(
            "delete",
            f"/api/storage/quota/rules/{record['uuid']}",
        ) as resp:
            data = await resp.json()
        return await self.wait_job(data["job"]["uuid"])

    async def enable_quota(
        self,
        volume_id: VolumeID,
        enable: bool = True,
    ) -> AsyncJobResult:
        async with self.send_request(
            "patch",
            f"/api/storage/volumes/{volume_id}",
            data={
                "quota.enabled": enable,
            },
        ) as resp:
            data = await resp.json()
        return await self.wait_job(data["job"]["uuid"])

    async def get_quota_report(
        self,
        svm_id: StorageID,
        volume_id: VolumeID,
        qtree_name: str,
    ) -> QuotaUsage:
        async with self.send_request(
            "get",
            "/api/storage/quota/reports",
            params={
                "type": "tree",
                "svm.uuid": str(svm_id),
                "volume.uuid": str(volume_id),
                "qtree.name": qtree_name,
                "fields": "space",
            },
        ) as resp:
            data = await resp.json()
            records = data["records"]
            if data["num_records"] == 0:
                raise NetAppClientError(
                    f"Quota report not found for the volume {volume_id} and the qtree {qtree_name}"
                )
            return QuotaUsage(
                used_bytes=records[0]["space"]["used"]["total"],
                limit_bytes=records[0]["space"]["hard_limit"],
            )

    async def get_qos_policies(self) -> List[Mapping[str, Any]]:
        async with self.send_request(
            "get",
            "/api/storage/qos/policies",
        ) as resp:
            data = await resp.json()
            qos_policies_metadata = data["records"]
            qos_policies = []
            for qos in qos_policies_metadata:
                policy = await self.get_qos_by_uuid(qos["uuid"])
                qos_policies.append(policy)
        return qos_policies

    async def get_qos_by_uuid(self, qos_uuid) -> Mapping[str, Any]:
        async with self.send_request(
            "get",
            f"/api/storage/qos/policies/{qos_uuid}",
        ) as resp:
            data = await resp.json()
            fixed = data["fixed"]
            qos_policy = {
                "uuid": data["uuid"],
                "name": data["name"],
                "fixed": {
                    "max_throughput_iops": fixed.get("max_throughput_iops", 0),
                    "max_throughput_mbps": fixed.get("max_throughput_mbps", 0),
                    "min_throughput_iops": fixed.get("min_throughput_iops", 0),
                    "min_throughput_mbps": fixed.get("min_throughput_mbps", 0),
                    "capacity_shared": fixed["capacity_shared"],
                },
                "svm": data["svm"],
            }
            return qos_policy

    async def get_qos_by_volume_id(self, volume_uuid) -> Mapping[str, Any]:
        async with self.send_request(
            "get",
            f"/api/storage/volumes/{volume_uuid}?fields=qos",
        ) as resp:
            data = await resp.json()
        return data["qos"]



from __future__ import annotations

import enum
import json
import os
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, AsyncIterator, Dict, List, Mapping

import aiohttp


class QuotaTypes(enum.StrEnum):
    DIRECTORY = "directory"
    USER = "user"
    GROUP = "group"


@dataclass
class QuotaThresholds:
    hard: int
    soft: int


class OneFSClient:
    endpoint: str
    user: str
    password: str
    api_version: str
    system_name: str
    _session: aiohttp.ClientSession

    def __init__(
        self,
        endpoint: str,
        user: str,
        password: str,
        *,
        api_version: str = "12",
        system_name: str = "nfs",
    ) -> None:
        self.endpoint = endpoint
        self.user = user
        self.password = password
        self.api_version = api_version
        self.system_name = system_name
        self._session = aiohttp.ClientSession()
        self._request_opts = {
            "auth": aiohttp.BasicAuth(self.user, self.password),
            "ssl": False,
            "raise_for_status": True,
        }

    async def aclose(self) -> None:
        await self._session.close()

    async def get_metadata(self) -> Mapping[str, Any]:
        cluster_metadata = await self.get_cluster_metadata()
        node_metadata = await self.get_node_metadata()
        volume_cluster = {
            "cluster": json.dumps(cluster_metadata),
            "nodes": json.dumps(node_metadata),
        }
        return volume_cluster

    @asynccontextmanager
    async def _request(
        self,
        method: str,
        path: str,
        **kwargs,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        async with self._session.request(
            method,
            f"{self.endpoint}/platform/{self.api_version}/{path}",
            auth=aiohttp.BasicAuth(self.user, self.password),
            ssl=False,
            raise_for_status=True,
            **kwargs,
        ) as resp:
            yield resp

    async def get_usage(self) -> Mapping[str, Any]:
        async with self._request("GET", "storagepool/storagepools") as resp:
            data = await resp.json()
        return {
            "capacity_bytes": data["storagepools"][0]["usage"]["total_bytes"],
            "used_bytes": data["storagepools"][0]["usage"]["used_bytes"],
        }

    async def get_list_lnn(self) -> List[int]:
        async with self._request("GET", "storagepool/storagepools") as resp:
            data = await resp.json()
        return data["storagepools"][0]["lnns"]

    async def get_node_hardware_info_by_lnn(self, lnn) -> Mapping[str, Any]:
        async with self._request("GET", f"cluster/nodes/{lnn}/hardware") as resp:
            data = await resp.json()
            node = data["nodes"][0]
        return {
            "id": node["id"],
            "model": node["product"],
            "configuration": node["configuration_id"],
            "serial_number": node["serial_number"],
        }

    async def get_node_status_by_lnn(self, lnn) -> Mapping[str, Any]:
        async with self._request("GET", f"cluster/nodes/{lnn}/status/nvram") as resp:
            data = await resp.json()
            node = data["nodes"][0]
        return {
            "batteries": node["batteries"],
            # "capacity": node["capacity"],
        }

    async def get_cluster_metadata(self) -> List[Dict[str, Any]]:
        try:
            cluster_metadata = []
            cluster_metadata.append({
                "config": await self.get_cluster_config(),
                "interface": await self.get_cluster_interface(),
            })
            return cluster_metadata
        except Exception as e:
            raise (e)

    async def get_cluster_config(self) -> Mapping[str, Any]:
        async with self._request("GET", "cluster/config") as resp:
            data = await resp.json()
        return {
            "name": data["name"],
            "release": data["onefs_version"]["release"],
            "build": data["onefs_version"]["build"],
            "local_lnn": data["local_lnn"],
        }

    async def get_cluster_interface(self) -> List[Mapping[str, Any]]:
        async with self._request("GET", "network/interfaces") as resp:
            data = await resp.json()
        return data["interfaces"]

    async def get_node_metadata(self) -> List[Dict[str, Mapping[str, Any]]]:
        try:
            lnns = await self.get_list_lnn()
            node_metadata = []
            for lnn in lnns:
                node_metadata.append({
                    "hardware": await self.get_node_hardware_info_by_lnn(lnn),
                    "status": await self.get_node_status_by_lnn(lnn),
                })
            return node_metadata
        except Exception as e:
            raise (e)

    async def get_drive_stats(self) -> Mapping[int, Any]:
        async with self._request("GET", "statistics/summary/drive") as resp:
            data = await resp.json()
        return data["drive"]

    async def get_protocol_stats(self) -> Mapping[str, Any]:
        async with self._request("GET", "statistics/summary/protocol-stats") as resp:
            data = await resp.json()
        return data["protocol-stats"]

    async def get_workload_stats(self) -> Mapping[str, Any]:
        async with self._request(
            "GET",
            "statistics/summary/workload",
            params={"system_names": self.system_name},
        ) as resp:
            data = await resp.json()
            return data["workload"][0]

    async def get_system_stats(self) -> Mapping[int, Any]:
        async with self._request("GET", "statistics/summary/system") as resp:
            data = await resp.json()
        return data["system"][0]

    async def list_all_quota(self) -> Mapping[str, Any]:
        async with self._request("GET", "quota/quotas") as resp:
            data = await resp.json()
        return data["quotas"]

    async def get_quota(self, quota_id: str) -> Mapping[str, Any]:
        async with self._request("GET", f"quota/quotas/{quota_id}") as resp:
            data = await resp.json()
        return data["quotas"][0]

    async def create_quota(
        self,
        path: os.PathLike,
        type_: QuotaTypes,
        thresholds: QuotaThresholds,
    ) -> Mapping[str, Any]:
        data = {
            "path": os.fsdecode(path),
            "type": type_.value,
            "include_snapshots": False,
            "thresholds": thresholds,
            "thresholds_on": "fslogicalsize",
            "enforced": True,
        }
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/hal+json",
        }
        async with self._request(
            "POST",
            "quota/quotas",
            headers=headers,
            data=data,
        ) as resp:
            msg = await resp.json()
        return msg

    async def delete_quota(self, quota_id):
        async with self._request(
            "DELETE",
            f"quota/quotas/{quota_id}",
        ) as resp:
            msg = await resp.json()
        return msg

    async def update_quota(self, quota_id: str, thresholds: QuotaThresholds):
        data = {
            "thresholds": thresholds,
        }
        async with self._request(
            "PUT",
            f"quota/quotas/{quota_id}",
            data=data,
        ) as resp:
            msg = await resp.json()
        return msg



import asyncio
import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Any, Final, FrozenSet, Literal, Mapping, Optional

import aiofiles
import aiofiles.os

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events import EventDispatcher, EventProducer
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import HardwareMetadata, QuotaConfig, QuotaScopeID

from ..abc import CAP_FAST_FS_SIZE, CAP_FAST_SIZE, CAP_METRIC, CAP_QUOTA, CAP_VFOLDER
from ..exception import (
    ExternalError,
    InvalidQuotaConfig,
    QuotaScopeAlreadyExists,
    QuotaScopeNotFoundError,
    StorageProxyError,
)
from ..types import CapacityUsage, FSPerfMetric, QuotaUsage
from ..vfs import BaseQuotaModel, BaseVolume
from .exceptions import VASTInvalidParameterError, VASTNotFoundError, VASTUnknownError
from .vastdata_client import VASTAPIClient, VASTQuotaID

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


VAST_QUOTA_ID_FILE_NAME: Final = ".vast-quota-id"


class VASTQuotaModel(BaseQuotaModel):
    def __init__(
        self,
        mount_path: Path,
        api_client: VASTAPIClient,
    ) -> None:
        super().__init__(mount_path)
        self.api_client = api_client

    async def _get_vast_quota_id(self, quota_scope_id: QuotaScopeID) -> VASTQuotaID | None:
        qs_path = self.mangle_qspath(quota_scope_id)

        def _read():
            try:
                with open(qs_path / VAST_QUOTA_ID_FILE_NAME, "r") as f:
                    return VASTQuotaID(f.read())
            except FileNotFoundError:
                return None

        return await asyncio.get_running_loop().run_in_executor(None, _read)

    async def _set_vast_quota_id(
        self, quota_scope_id: QuotaScopeID, vast_quota_id: VASTQuotaID
    ) -> None:
        qs_path = self.mangle_qspath(quota_scope_id)

        def _write():
            qs_path.mkdir(parents=True, exist_ok=True)
            with open(qs_path / VAST_QUOTA_ID_FILE_NAME, "w") as f:
                f.write(str(vast_quota_id))

        await asyncio.get_running_loop().run_in_executor(None, _write)

    async def _rm_vast_quota_id(self, quota_scope_id: QuotaScopeID) -> None:
        qs_path = self.mangle_qspath(quota_scope_id)

        try:
            await aiofiles.os.remove(qs_path / VAST_QUOTA_ID_FILE_NAME)
        except FileNotFoundError:
            log.warning(f"vast quota id file not found (qid: {quota_scope_id}). skip")

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        try:
            await aiofiles.os.makedirs(qspath)
        except FileExistsError:
            vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
            if vast_quota_id is not None:
                quota = await self.api_client.get_quota(vast_quota_id)
                if quota is not None:
                    raise QuotaScopeAlreadyExists
        if options is not None:
            try:
                quota = await self.api_client.set_quota(
                    qspath,
                    soft_limit=options.limit_bytes,
                    hard_limit=options.limit_bytes,
                )
            except VASTInvalidParameterError:
                raise InvalidQuotaConfig
            except VASTUnknownError as e:
                raise ExternalError(str(e))
            await self._set_vast_quota_id(quota_scope_id, quota.id)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
        if vast_quota_id is None:
            raise QuotaScopeNotFoundError
        try:
            await self.api_client.modify_quota(
                vast_quota_id,
                soft_limit=config.limit_bytes,
                hard_limit=config.limit_bytes,
            )
        except VASTInvalidParameterError:
            raise InvalidQuotaConfig
        except VASTUnknownError as e:
            raise ExternalError(str(e))

    async def describe_quota_scope(self, quota_scope_id: QuotaScopeID) -> Optional[QuotaUsage]:
        qspath = self.mangle_qspath(quota_scope_id)
        if not qspath.exists():
            return None
        if (vast_quota_id := await self._get_vast_quota_id(quota_scope_id)) is None:
            return None
        if (quota := await self.api_client.get_quota(vast_quota_id)) is None:
            return None
        return QuotaUsage(
            used_bytes=quota.used_capacity,
            limit_bytes=quota.hard_limit,
        )

    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        vast_quota_id = await self._get_vast_quota_id(quota_scope_id)
        if vast_quota_id is None:
            raise QuotaScopeNotFoundError
        try:
            await self.api_client.remove_quota(vast_quota_id)
        except VASTNotFoundError:
            raise QuotaScopeNotFoundError
        await self._rm_vast_quota_id(quota_scope_id)

    async def delete_quota_scope(self, quota_scope_id: QuotaScopeID) -> None:
        await self.unset_quota(quota_scope_id)
        qspath = self.mangle_qspath(quota_scope_id)
        await aiofiles.os.rmdir(qspath)


class VASTVolume(BaseVolume):
    api_client: VASTAPIClient

    name = "vast"

    def __init__(
        self,
        local_config: Mapping[str, Any],
        mount_path: Path,
        *,
        etcd: AsyncEtcd,
        event_dispathcer: EventDispatcher,
        event_producer: EventProducer,
        options: Optional[Mapping[str, Any]] = None,
    ) -> None:
        super().__init__(
            local_config,
            mount_path,
            etcd=etcd,
            options=options,
            event_dispathcer=event_dispathcer,
            event_producer=event_producer,
        )
        ssl_verify = self.config.get("vast_verify_ssl", False)
        self.api_client = VASTAPIClient(
            self.config["vast_endpoint"],
            self.config["vast_username"],
            self.config["vast_password"],
            storage_base_dir=self.config["vast_storage_base_dir"],
            api_version=self.config["vast_api_version"],
            ssl=ssl_verify,
        )

    async def shutdown(self) -> None:
        self.api_client.cache.cluster_info = None

    async def create_quota_model(self) -> VASTQuotaModel:
        return VASTQuotaModel(self.mount_path, self.api_client)

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER, CAP_METRIC, CAP_QUOTA, CAP_FAST_FS_SIZE, CAP_FAST_SIZE])

    async def get_hwinfo(self) -> HardwareMetadata:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return {
                "status": "unavailable",
                "status_info": None,
                "metadata": {},
            }
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        healthy_status: Literal["healthy", "degraded", "unavailable"] = "unavailable"
        match clsuter_info.state.lower():
            case "online" | "healthy":
                healthy_status = "healthy"
            case "init":
                healthy_status = "degraded"
            case "unknown":
                healthy_status = "unavailable"
        quotas = await self.api_client.list_quotas()
        return {
            "status": healthy_status,
            "status_info": clsuter_info.state,
            "metadata": {
                "quota": json.dumps([asdict(q) for q in quotas]),
                "cluster_info": json.dumps(asdict(clsuter_info)),
            },
        }

    async def get_performance_metric(self) -> FSPerfMetric:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return FSPerfMetric(
                iops_read=-1,
                iops_write=-1,
                io_bytes_read=-1,
                io_bytes_write=-1,
                io_usec_read=-1,
                io_usec_write=-1,
            )
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        return FSPerfMetric(
            iops_read=clsuter_info.rd_iops,
            iops_write=clsuter_info.wr_iops,
            io_usec_read=clsuter_info.rd_latency,
            io_usec_write=clsuter_info.wr_latency,
            io_bytes_read=clsuter_info.rd_bw,
            io_bytes_write=clsuter_info.wr_bw,
        )

    async def get_fs_usage(self) -> CapacityUsage:
        cluster_id: int = self.config["vast_cluster_id"]
        try:
            clsuter_info = await self.api_client.get_cluster_info(cluster_id)
        except VASTUnknownError:
            return CapacityUsage(
                used_bytes=-1,
                capacity_bytes=-1,
            )
        if clsuter_info is None:
            raise StorageProxyError(f"vast cluster not found. (id: {cluster_id})")
        return CapacityUsage(
            used_bytes=clsuter_info.physical_space_in_use,
            capacity_bytes=clsuter_info.physical_space,
        )



from __future__ import annotations

import logging
from contextlib import asynccontextmanager as actxmgr
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Final,
    Mapping,
    Optional,
    Type,
)

import aiohttp_cors
from aiohttp import web

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events import (
    EventDispatcher,
    EventProducer,
)
from ai.backend.common.logging import BraceStyleAdapter

from .abc import AbstractVolume
from .api.client import init_client_app
from .api.manager import init_manager_app
from .api.types import WebMiddleware
from .cephfs import CephFSVolume
from .ddn import EXAScalerFSVolume
from .dellemc import DellEMCOneFSVolume
from .exception import InvalidVolumeError
from .gpfs import GPFSVolume
from .netapp import NetAppVolume
from .plugin import (
    BasePluginContext,
    StorageClientWebappPluginContext,
    StorageManagerWebappPluginContext,
    StoragePluginContext,
)
from .purestorage import FlashBladeVolume
from .types import VolumeInfo
from .vast import VASTVolume
from .vfs import BaseVolume
from .watcher import WatcherClient
from .weka import WekaVolume
from .xfs import XfsVolume

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "storage-proxy"

DEFAULT_BACKENDS: Mapping[str, Type[AbstractVolume]] = {
    FlashBladeVolume.name: FlashBladeVolume,
    BaseVolume.name: BaseVolume,
    XfsVolume.name: XfsVolume,
    NetAppVolume.name: NetAppVolume,
    # NOTE: Dell EMC has two different storage: PowerStore and PowerScale (OneFS).
    #       We support the latter only for now.
    DellEMCOneFSVolume.name: DellEMCOneFSVolume,
    WekaVolume.name: WekaVolume,
    GPFSVolume.name: GPFSVolume,  # IBM SpectrumScale or GPFS
    "spectrumscale": GPFSVolume,  # IBM SpectrumScale or GPFS
    CephFSVolume.name: CephFSVolume,
    VASTVolume.name: VASTVolume,
    EXAScalerFSVolume.name: EXAScalerFSVolume,
}


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: list[WebMiddleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["ctx"] = root_app["ctx"]

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


class RootContext:
    volumes: dict[str, AbstractVolume]
    pid: int
    etcd: AsyncEtcd
    local_config: Mapping[str, Any]
    dsn: str | None
    event_producer: EventProducer
    event_dispatcher: EventDispatcher
    watcher: WatcherClient | None

    def __init__(
        self,
        pid: int,
        pidx: int,
        node_id: str,
        local_config: Mapping[str, Any],
        etcd: AsyncEtcd,
        *,
        event_producer: EventProducer,
        event_dispatcher: EventDispatcher,
        watcher: WatcherClient | None,
        dsn: Optional[str] = None,
    ) -> None:
        self.volumes = {}
        self.pid = pid
        self.pidx = pidx
        self.node_id = node_id
        self.etcd = etcd
        self.local_config = local_config
        self.dsn = dsn
        self.event_producer = event_producer
        self.event_dispatcher = event_dispatcher
        self.watcher = watcher
        self.cors_options = {
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=False, expose_headers="*", allow_headers="*"
            ),
        }

    async def __aenter__(self) -> None:
        self.client_api_app = await init_client_app(self)
        self.manager_api_app = await init_manager_app(self)
        self.backends = {
            **DEFAULT_BACKENDS,
        }
        await self.init_storage_plugin()
        self.manager_webapp_plugin_ctx = await self.init_storage_webapp_plugin(
            StorageManagerWebappPluginContext(self.etcd, self.local_config),
            self.manager_api_app,
        )
        self.client_webapp_plugin_ctx = await self.init_storage_webapp_plugin(
            StorageClientWebappPluginContext(self.etcd, self.local_config),
            self.client_api_app,
        )

    async def init_storage_plugin(self) -> None:
        plugin_ctx = StoragePluginContext(self.etcd, self.local_config)
        await plugin_ctx.init()
        self.storage_plugin_ctx = plugin_ctx
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            log.info("Loading storage plugin: {0}", plugin_name)
            volume_cls = plugin_instance.get_volume_class()
            self.backends[plugin_name] = volume_cls

    async def init_storage_webapp_plugin(
        self, plugin_ctx: BasePluginContext, root_app: web.Application
    ) -> BasePluginContext:
        await plugin_ctx.init()
        for plugin_name, plugin_instance in plugin_ctx.plugins.items():
            if self.pid == 0:
                log.info("Loading storage webapp plugin: {0}", plugin_name)
            subapp, global_middlewares = await plugin_instance.create_app(self.cors_options)
            _init_subapp(plugin_name, root_app, subapp, global_middlewares)
        return plugin_ctx

    def list_volumes(self) -> Mapping[str, VolumeInfo]:
        return {name: VolumeInfo(**info) for name, info in self.local_config["volume"].items()}

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        for volume in self.volumes.values():
            await volume.shutdown()

        await self.storage_plugin_ctx.cleanup()
        await self.manager_webapp_plugin_ctx.cleanup()
        await self.client_webapp_plugin_ctx.cleanup()
        return None

    @actxmgr
    async def get_volume(self, name: str) -> AsyncIterator[AbstractVolume]:
        if name in self.volumes:
            yield self.volumes[name]
        else:
            try:
                volume_config = self.local_config["volume"][name]
            except KeyError:
                raise InvalidVolumeError(name)
            volume_cls: Type[AbstractVolume] = self.backends[volume_config["backend"]]
            volume_obj = volume_cls(
                local_config=self.local_config,
                mount_path=Path(volume_config["path"]),
                options=volume_config["options"] or {},
                etcd=self.etcd,
                event_dispathcer=self.event_dispatcher,
                event_producer=self.event_producer,
            )

            await volume_obj.init()
            self.volumes[name] = volume_obj

            yield volume_obj



from enum import StrEnum
from typing import Final

import trafaret as t

from ai.backend.common import validators as tx

DEFAULT_CLUSTER_ID: Final = 1


class APIVersion(StrEnum):
    V1 = "v1"
    V2 = "v2"
    LATEST = "latest"


config_iv = t.Dict({
    t.Key("vast_endpoint"): t.String(),
    t.Key("vast_username"): t.String(),
    t.Key("vast_password"): t.String(),
    t.Key("vast_verify_ssl", default=False): t.Bool(),
    t.Key("vast_api_version", default=APIVersion.V2): tx.Enum(APIVersion),
    t.Key("vast_cluster_id", default=DEFAULT_CLUSTER_ID): t.Int,
    t.Key("vast_storage_base_dir", default="/"): t.String(),
}).allow_extra("*")



import json
from typing import Any, Optional

from aiohttp import web
from aiohttp.web_exceptions import HTTPNotFound


class StorageProxyError(Exception):
    pass


class ExecutionError(StorageProxyError):
    pass


class ExternalError(StorageProxyError):
    pass


class NotEmptyError(StorageProxyError):
    pass


class VFolderCreationError(StorageProxyError):
    pass


class VFolderNotFoundError(StorageProxyError):
    pass


class QuotaScopeNotFoundError(StorageProxyError, HTTPNotFound):
    pass


class QuotaScopeAlreadyExists(StorageProxyError):
    pass


class InvalidQuotaConfig(StorageProxyError):
    pass


class InvalidSubpathError(StorageProxyError):
    pass


class InvalidQuotaScopeError(StorageProxyError):
    pass


class InvalidVolumeError(StorageProxyError):
    pass


class WatcherClientError(RuntimeError):
    pass


class InvalidAPIParameters(web.HTTPBadRequest):
    def __init__(
        self,
        type_suffix: str = "invalid-api-params",
        title: str = "Invalid API parameters",
        msg: Optional[str] = None,
        data: Any = None,
    ) -> None:
        payload = {
            "type": f"https://api.backend.ai/probs/storage/{type_suffix}",
            "title": title,
        }
        if msg is not None:
            payload["title"] = f"{title} ({msg})"
        if data is not None:
            payload["data"] = data
        super().__init__(
            text=json.dumps(payload),
            content_type="application/problem+json",
        )



import os
import sys
from pathlib import Path
from pprint import pformat
from typing import Any

import trafaret as t

from ai.backend.common import validators as tx
from ai.backend.common.config import (
    ConfigurationError,
    check,
    etcd_config_iv,
    override_key,
    override_with_env,
    read_from_file,
)
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.logging import logging_config_iv

from .types import VolumeInfo

_max_cpu_count = os.cpu_count()
try:
    _file_perm = (Path(__file__).parent / "server.py").stat()
    _default_uid = _file_perm.st_uid
    _default_gid = _file_perm.st_gid
except IOError:
    _default_uid = os.getuid()
    _default_gid = os.getgid()


local_config_iv = (
    t.Dict(
        {
            t.Key("storage-proxy"): t.Dict(
                {
                    t.Key("ipc-base-path", default="/tmp/backend.ai/ipc"): tx.Path(
                        type="dir", auto_create=True
                    ),
                    t.Key("node-id"): t.String,
                    t.Key("num-proc", default=_max_cpu_count): t.Int[1:_max_cpu_count],
                    t.Key("pid-file", default=os.devnull): tx.Path(
                        type="file",
                        allow_nonexisting=True,
                        allow_devnull=True,
                    ),
                    t.Key("event-loop", default="asyncio"): t.Enum("asyncio", "uvloop"),
                    t.Key("scandir-limit", default=1000): t.Int[0:],
                    t.Key("max-upload-size", default="100g"): tx.BinarySize,
                    t.Key("secret"): t.String,  # used to generate JWT tokens
                    t.Key("session-expire"): tx.TimeDuration,
                    t.Key("user", default=None): tx.UserID(
                        default_uid=_default_uid,
                    ),
                    t.Key("group", default=None): tx.GroupID(
                        default_gid=_default_gid,
                    ),
                    tx.AliasedKey(
                        ["aiomonitor-termui-port", "aiomonitor-port"], default=48300
                    ): t.ToInt[1:65535],
                    t.Key("aiomonitor-webui-port", default=49300): t.ToInt[1:65535],
                    t.Key("watcher-insock-path-prefix", default=None): t.Null
                    | t.String(allow_blank=False),
                    t.Key("watcher-outsock-path-prefix", default=None): t.Null
                    | t.String(allow_blank=False),
                    t.Key("use-watcher", default=False): t.Bool(),
                    t.Key("use-experimental-redis-event-dispatcher", default=False): t.ToBool,
                },
            ),
            t.Key("logging"): logging_config_iv,
            t.Key("api"): t.Dict(
                {
                    t.Key("client"): t.Dict(
                        {
                            t.Key("service-addr"): tx.HostPortPair(
                                allow_blank_host=True,
                            ),
                            t.Key("ssl-enabled"): t.ToBool,
                            t.Key("ssl-cert", default=None): t.Null | tx.Path(type="file"),
                            t.Key("ssl-privkey", default=None): t.Null | tx.Path(type="file"),
                        },
                    ),
                    t.Key("manager"): t.Dict(
                        {
                            t.Key("service-addr"): tx.HostPortPair(
                                allow_blank_host=True,
                            ),
                            t.Key("ssl-enabled"): t.ToBool,
                            t.Key("ssl-cert", default=None): t.Null | tx.Path(type="file"),
                            t.Key("ssl-privkey", default=None): t.Null | tx.Path(type="file"),
                            t.Key("secret"): t.String,  # used to authenticate managers
                        },
                    ),
                },
            ),
            t.Key("volume"): t.Mapping(
                t.String,
                VolumeInfo.as_trafaret(),  # volume name -> details
            ),
            t.Key("debug"): t.Dict(
                {
                    t.Key("enabled", default=False): t.ToBool,
                    t.Key("asyncio", default=False): t.ToBool,
                    t.Key("enhanced-aiomonitor-task-info", default=False): t.ToBool,
                    t.Key("log-events", default=False): t.ToBool,
                },
            ).allow_extra("*"),
        },
    )
    .merge(etcd_config_iv)
    .allow_extra("*")
)


def load_local_config(config_path: Path | None, debug: bool = False) -> dict[str, Any]:
    # Determine where to read configuration.
    raw_cfg, cfg_src_path = read_from_file(config_path, "storage-proxy")
    os.chdir(cfg_src_path.parent)

    override_with_env(raw_cfg, ("etcd", "namespace"), "BACKEND_NAMESPACE")
    override_with_env(raw_cfg, ("etcd", "addr"), "BACKEND_ETCD_ADDR")
    override_with_env(raw_cfg, ("etcd", "user"), "BACKEND_ETCD_USER")
    override_with_env(raw_cfg, ("etcd", "password"), "BACKEND_ETCD_PASSWORD")
    if debug:
        override_key(raw_cfg, ("debug", "enabled"), True)

    try:
        local_config = check(raw_cfg, local_config_iv)
        local_config["_src"] = cfg_src_path
        return local_config
    except ConfigurationError as e:
        print(
            "ConfigurationError: Validation of storage-proxy local config has failed:",
            file=sys.stderr,
        )
        print(pformat(e.invalid_data), file=sys.stderr)
        raise


def load_shared_config(local_config: dict[str, Any]) -> AsyncEtcd:
    etcd_credentials = None
    if local_config["etcd"]["user"]:
        etcd_credentials = {
            "user": local_config["etcd"]["user"],
            "password": local_config["etcd"]["password"],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        ConfigScopes.NODE: f"nodes/storage/{local_config['storage-proxy']['node-id']}",
    }
    etcd = AsyncEtcd(
        local_config["etcd"]["addr"],
        local_config["etcd"]["namespace"],
        scope_prefix_map,
        credentials=etcd_credentials,
    )
    return etcd
